"""
This folder contains 3rd party packages that are used in this Weasel project.
You can find a list of these below with their respective version. 
Each one has a folder allocated, which contains their own README.md file with their licensing details (all open-source)

- dcm4che Version 5.23.1

"""