# setup.py
# Dummy file
# This file is there to fix a limitation in setuptools
# https://realpython.com/pypi-publish-python-package/#prepare-your-package-for-publication

from setuptools import setup

setup()